;*****************************************************************************************
;
;       PROGRAM:                A/V MODULATOR + PLL 4 FREQ. ATV
;
;       ASSEMBLER:              MPASMWIN v02.30
;
;       Created by:		P. TURRINI             pierbassano.turrini@fastwebnet.it
;
;       REVISION HISTORY:
;
;       VERSION:        DATE            COMMENTS
;       1.00            24-05-2004      FIRST RELEASE.
;       
;       DESCRIPTION:
;
;       PROGRAMMABLE SYNTHESIZER USING A PIC 16F84 AND A TSA5511 PLL.
;       RC CLOCK FOR PIC (22PF & 5.6K = APPROX 3.6MHZ, = 900KHZ INSTRUCTION CYCLE TIME):
;
;       PLL FREQUENCY FOR ATV.
;       FREQUENCY STEPS OF 50KHZ WITH A CRYSTAL REFERENCE OF 3.2MHZ.
;       FREQUENCY STEPS OF 62.5KHZ WITH A CRYSTAL REFERENCE OF 4MHZ.
;
;	TO MIMIMISE DISTURBANCE TO THE PLL, THE DIVIDER INFO THAT SETS THE PLL, IS SENT
;       BY THE PIC TO THE PLL ONLY IF THE FREQUENCY NEEDS TO BE CHANGED.
;
;       WHEN THE PLL FREQUENCY IS CHANGED, IT IS ASSUMED THAT THE PLL IS UNLOCKED. THE PIC
;       CHECKS THE PLL STATUS CONTINUOUSLY UNTIL THE PLL LOCKS. ONCE LOCKED, STATUS CHECKS
;       ARE MADE EVERY SECOND. IF THE PLL GOES OUT OF LOCK, CONTINUOUS CHECKS ARE STARTED
;       AGAIN. (TO DISABLE THE CHECKS EVERY SECOND, CHANGE THE 'LOCK_TM' EQUATE TO 255)
;
;*****************************************************************************************
;*****************************************************************************************
;       REGISTER DEFINITIONS
;*****************************************************************************************

W               EQU             0
F               EQU             1

INDF            SET             0X00
TMR0            SET             0X01
PCL             SET             0X02
STATUS          SET             0X03
FSR             SET             0X04
PORTA           SET             0X05
PORTB           SET             0X06
EEDATA          SET             0X08
EEADR           SET             0X09
INTCON          SET             0X0B
OPTION_REG      SET             0X01
EECON1          SET             0X08
EECON2          SET             0X09
#DEFINE         CARRY           STATUS,0
#DEFINE         ZERO            STATUS,2
#DEFINE         RP0             STATUS,5
#DEFINE         T0IF            INTCON,2
#DEFINE         RD              EECON1,0
#DEFINE         WR              EECON1,1
#DEFINE         WREN            EECON1,2
#DEFINE         WRERR           EECON1,3
#DEFINE         EEIF            EECON1,4

;       CONFIGURATION BITS

_CP_ON          EQU     H'3FEF'
_CP_OFF         EQU     H'3FFF'
_PWRTE_ON       EQU     H'3FFF'
_PWRTE_OFF      EQU     H'3FF7'
_WDT_OFF        EQU     H'3FFB'
_HS_OSC         EQU     H'3FFE'
_RC_OSC         EQU     H'3FFF'

;*****************************************************************************************
;       PROCESSOR DEFINITIONS
;*****************************************************************************************

        PROCESSOR       16F84                   ;SELECT PROCCESOR, AND ASSEMBLER OPTIONS.
;       INCLUDE		P16F84.INC
        LIST            F=INHX8M,N=0
        RADIX           DEC

 	__CONFIG        _RC_OSC & _WDT_OFF      ;RC OSCILLATOR BUT NO WATCHDOG.
        

;*****************************************************************************************
;               GENERAL DEFINITIONS
;*****************************************************************************************

#DEFINE         FREQ1	        PORTB,0         ;FREQ 1224 MHz.
#DEFINE         FREQ2	        PORTB,1         ;FREQ 1240 MHz.
#DEFINE         FREQ3	        PORTB,2         ;FREQ 1256 MHz.
#DEFINE         FREQ4	        PORTB,3         ;FREQ 1272 MHz.
#DEFINE         TX_EN           PORTA,2         ;TX ENABLE OUTPUT PIN. 0 = ENABLED.
#DEFINE         SDA             PORTA,1         ;I2C DATA LINE.
#DEFINE         SCL             PORTA,0         ;I2C CLOCK LINE.

#DEFINE         PLL_VALID       FLAGS,2         ;PLL DIVIDER NUMBER IS VALID.

;*****************************************************************************************
;               GENERAL EQUATES
;*****************************************************************************************

RESETVECTOR     EQU             H'00'           ;PIC16F84 RESET VECTOR.
INTVECTOR       EQU             H'04'           ;PIC16F84 INTERUPT VECTOR.

LOCK_TM         EQU             110             ;TIME BETWEEN STATUS CHECKS. (100 x 10mS)
                                                ;255 = NO CHECKS ONCE IN LOCK.

PLL_ADDRESS     EQU             B'11000010'     ;PLL CHIP ADDRESS.

; FOR 4 MHZ CRYSTAL REFERENCE

F1224_H   EQU       B'01001100'
F1224_L   EQU       B'10000000'

F1240_H   EQU       B'01001101'
F1240_L   EQU       B'10000000'

F1256_H   EQU       B'01001110'
F1256_L   EQU       B'10000000'

F1272_H   EQU       B'01001111'
F1272_L   EQU       B'10000000'

; FOR 3.2 MHZ CRYSTAL REFERENCE

;F1224_H   EQU       B'01011111'
;F1224_L   EQU       B'11000000'

;F1240_H   EQU       B'01100000'
;F1240_L   EQU       B'11100000'

;F1256_H   EQU       B'01100010'
;F1256_L   EQU       B'00100000'

;F1272_H   EQU       B'01100011'
;F1272_L   EQU       B'01100000'


;BYTE4     EQU       B'10101110'         	;byte 4 5511
;BYTE5     EQU       B'11111111'         	;byte 5 5511

;*****************************************************************************************
;               RAM VARIABLES 68 BYTES MAX
;*****************************************************************************************

                CBLOCK  H'0C'                   ;RAM STARTS AT H'0C'

                COUNT                           ;GENERAL COUNTER VARIABLES.
                COUNT1
                COUNT2
                SAVE_STATUS                     ;FOR INTERUPT SERVICE ROUTINE.
                SAVE_W_REG                      ;             "
                LOCK_TIMER                      ;TIMER FOR CHECKING PLL STATUS.
                FLAGS                           ;8 MISC FLAGS.
                PLL_STATUS                      ;STATUS BYTE FROM PLL READ.
                TEMP                            ;TEMPORARY STORAGE.
                DIV_H                           ;DIVIDER NUMBER MSB.
                DIV_L                           ;DIVIDER NUMBER LSB.
                TPORTB				;STATUS PORTB
                
                ENDC


;*****************************************************************************************
;
;               INITIALISE THE HARDWARE.
;
;*****************************************************************************************

                ORG     RESETVECTOR

                GOTO    CONTINUE                ;JUMP OVER INTERUPT ROUTINES.

;*****************************************************************************************
;               INTERRUPT SERVICE ROUTINE
;
;               SERVICE THE TMR0 INTERUPT (EVERY 10mS)
;
;               DEC THE 8 BIT LOCK TIMER, DOWN COUNTER. WONT DEC PAST 0.
;               WHEN EQUAL TO 0 ITS TIME TO CHECK IF THE PLL IS LOCKED.
;*****************************************************************************************

                ORG     INTVECTOR

                BTFSS   T0IF                    ;WAS THE INTERUPT CAUSED BY TMR0 ?
                RETFIE                          ;N. JUST RETURN & RE-ENABLE INTERUPTS.
                MOVWF   SAVE_W_REG              ;Y. SAVE WREG & STATUS REGS.
                SWAPF   STATUS,W
                MOVWF   SAVE_STATUS
                BCF     RP0                     ;MAKE SURE WE ARE ADDRESSING BANK 0.

;-----------------------------------------------------------------------------------------

		INCF    LOCK_TIMER,W            ;GET THE LOCK TIME COUNTER.
		BTFSC   ZERO                    ;IS IT = 255?
		GOTO    RS_INT_STATUS
		MOVF    LOCK_TIMER,W            ;N. GET THE LOCK TIME COUNTER.
		BTFSS   ZERO                    ;IS IT = 0?
		DECF    LOCK_TIMER,F            ;N. DEC THE COUNTER.
			                        ;Y. DONT DEC THE COUNTER.

;-----------------------------------------------------------------------------------------

RS_INT_STATUS:  SWAPF   SAVE_STATUS,W
                MOVWF   STATUS                  ;RESTORE STATUS REG.
                SWAPF   SAVE_W_REG,F            ;SAVE WREG.
                SWAPF   SAVE_W_REG,W            ;RESTORE WREG.
                BCF     T0IF                    ;CLEAR TMR0 INTERUPT FLAG.
                RETFIE
		RETURN
		
;*****************************************************************************************
;
;               INITIALISE THE HARDWARE. (CONTINUED)
;
;*****************************************************************************************

CONTINUE:	CALL	LONG_DELAY		;WAIT FOR POWER_ON 
		BSF     RP0                     ;SELECT PAGE 1 FOR TRIS REG ACCESS.
                MOVLW   B'00000010'
                MOVWF   PORTA                   ;RA0,RA2 = OUTPUT, RA1 = INPUT.
                MOVLW   B'11111111'
                MOVWF   PORTB                   ;RB0-RB7 = INPUT.
                MOVLW   B'00000100'             ;SET PRESCALER TO TMR0.
                MOVWF   OPTION_REG              ;TMR0 RATE = 1/32 OF INSTRUCTION CLOCK.
                BCF     RP0                     ;RETURN TO PAGE 0 FOR PORT ACCESS.

                CLRF    TMR0                    ;SETS 1ST INTERUPT TO 10mS.
                MOVLW   B'10100000'             ;ENABLE TMR0 OVERFLOW INTERUPT.
                MOVWF   INTCON

                MOVLW   B'00000100'             ;SET RA1 LOW & TX_EN HIGHT
                MOVWF   PORTA                   ;WHEN SET AS AN OUTPUT.

                BSF     TX_EN                   ;SET TX_EN PIN HIGH, PLL NOT LOCKED YET.
		BCF	PLL_VALID		;SET THE VALID FLAG = INVALID
		
		
MAIN_LOOP:      MOVF	PORTB,W			;READ PORTB STATUS
		ANDLW	B'00001111'
		SUBWF	TPORTB,W		;AND VERIFY IF CHANGED
		BTFSS	ZERO
		BCF	PLL_VALID		;SET THE VALID FLAG = INVALID
		BTFSS   PLL_VALID               ;IS THE PLL DIVIDER SETTING VALID 
                CALL    KEY_SET			;N, SET FREQUENCY
                CALL	GET_PLL_STATUS		;READ PLL STATUS
                BTFSS   PLL_STATUS,6            ;AND, IS THE PLL LOCKED?
                CALL    PLL_NOT_LOCKED
                GOTO    MAIN_LOOP
                
PLL_NOT_LOCKED: CALL    UPDATE_PLL              ;UPDATE PLL AND RETURN THE DISPLAY FREQ.
                GOTO    MAIN_LOOP               ;N. CONTINUE

;*****************************************************************************************
;  MISC ROUTINES USED TO UPDATE PLL.
;
;                       UPDATE_PLL              SET THE PLL DIVIDER.
;                                               ALSO CHECKS PLL STATUS.
;*****************************************************************************************

UPDATE_PLL:     CALL    GET_PLL_STATUS          ;UPDATE THE STATUS BYTE. (PLL LOCKED ?)
                CALL    SET_PLL                 ;SEND THE DIVIDER NUMBER TO THE PLL.
                RETURN

;*****************************************************************************************
;  I2C ROUTINES.        SET_PLL                 GIVEN THE DIVIDER NUMBER IN DIV, CHECKS
;                                               THAT THE NUMBER IS VALID THEN SENDS IT
;                                               TO THE PLL.
;                       GET_PLL_STATUS          GET THE STATUS BYTE FROM THE PLL,
;                                               AND SETS THE TX_EN LEAD.
;                       I2C_START               SEND AN I2C START SIGNAL ON THE BUS.
;                       I2C_TX_BYTE             SEND THE BYTE IN W TO THE BUS.
;                       I2C_STOP                SEND AN I2C STOP SIGNAL ON THE BUS.
;*****************************************************************************************

SET_PLL:        BSF     PLL_VALID               ;SET THE VALID FLAG.
                MOVF	PORTB,W	                ;READ PORTB STATUS.
                ANDLW	B'00001111'
                SUBWF	TPORTB,W		;AND VERIFY IF CHANGED
                BTFSC   ZERO
                GOTO    DONT_SEND               ;Y. NO NEED TO RE-SEND THE NUMBER.
                
WRT_PLL:        CALL    I2C_START
                MOVLW   PLL_ADDRESS             ;SEND THE PLL ADDRESS.
                CALL    I2C_TX_BYTE
                MOVF	DIV_H,W			;SEND MSD BYTE OF THE PLL DIVIDER.
                CALL    I2C_TX_BYTE
                MOVF	DIV_L,W			;SEND LSB BYTE OF THE PLL DIVIDER.
                CALL    I2C_TX_BYTE
                MOVLW   B'10101110'		;SET CHARGE PUMP CURRENT TO 50uA,
                CALL    I2C_TX_BYTE             ;P6=REF, P7=F/DIV.
                MOVLW   B'11111111'		;OUTPUT PINS ACTIVE.
                CALL    I2C_TX_BYTE
                CALL    I2C_STOP                ;SEND THE STOP COMMAND.

                BCF     PLL_STATUS,6            ;THE PLL IS NOT LOCKED NOW.
                				;THE DIVIDER NUMBER WAS VALID.
		CALL	LONG_DELAY		;WAIT 100mS TO INSURE PLL LOCK
DONT_SEND:      RETURN                          

;-----------------------------------------------------------------------------------------

GET_PLL_STATUS: BTFSS   PLL_STATUS,6            ;WAS PLL LOCKED WHEN LAST CHECKED?
                GOTO    GET_STATUS              ;N. GET THE CURRENT STATUS.
                MOVF    LOCK_TIMER,W            ;GET THE CURRENT PLL LOCKED TIMER.
                BTFSS   ZERO                    ;HAS THE COUNTER REACHED 0 ?
                GOTO    UPDATE_TX_EN            ;N. JUST UPDATE THE TX ENABLE PIN.
                CALL    RS_LOCKED_TMR           ;Y. RESET THE LOCKED TIMER.


GET_STATUS:     CALL    I2C_START               ;SEND AN I2C START SIGNAL.
                MOVLW   PLL_ADDRESS + 1         ;I2C ADDRESS & READ BIT.
                CALL    I2C_TX_BYTE             ;SEND THE ADDRESS BYTE TO THE PLL.

                MOVLW   8
                MOVWF   COUNT                   ;SET THE NUMBER OF BITS TO RECEIVE.
RX_BIT_LOOP:    CALL    I2C_DELAY
                BSF     SCL
                CALL    I2C_DELAY
                RRF     PORTA,F                 ;MOVE THE SDA (BIT 0) INTO THE CARRY BIT.
                RRF     PORTA,W                 ;MOVE THE SDA (BIT 1) INTO THE CARRY BIT.
                RLF     PLL_STATUS,F            ;SAVE THE RECEIVED BIT.
                BCF     SCL
                DECFSZ  COUNT,F
                GOTO    RX_BIT_LOOP             ;LOOP UNTIL 8 BITS RECEIVED.
                CALL    LET_SDA_HIGH            ;SEND AN I2C NEVATIVE ACKNOWLEDGE.
                CALL    I2C_CLOCK               ;(SDA HIGH)
                CALL    I2C_STOP                ;SEND AN I2C STOP SIGNAL.

                BTFSC   PLL_STATUS,6            ;IS PLL LOCKED NOW?
                CALL    RS_LOCKED_TMR           ;Y. RESET THE LOCKED TIMER.

UPDATE_TX_EN:   BTFSS   PLL_STATUS,6            ;IS THE PLL LOCKED?
                GOTO    NO_TX_EN
                BTFSS   PLL_VALID               ;Y. IS THE PLL DIVIDER SETING VALID?
                GOTO    NO_TX_EN
                BCF     TX_EN                   ;Y. ENABLE THE TX_EN LEAD.
                RETURN
NO_TX_EN:       BSF     TX_EN                   ;N. DONT ENABLE THE TX_EN LEAD.
                RETURN

RS_LOCKED_TMR:  MOVLW   LOCK_TM                 ;RESET THE PLL LOCKED TIMER.
                MOVWF   LOCK_TIMER
                RETURN

;-----------------------------------------------------------------------------------------
; I2C ROUTINE
;-----------------------------------------------------------------------------------------

I2C_TX_BYTE:    MOVWF   TEMP                    ;SAVE THE BYTE TO TX IN TEMP.
                MOVLW   8                       ;SET THE NUMBER OF BITS TO TRANSMIT.
                MOVWF   COUNT
TX_BIT_LOOP:    RLF     TEMP,F                  ;MOVE THE MSB INTO THE CARRY.
                BTFSC   CARRY
                CALL    LET_SDA_HIGH            ;IF CARRY IS SET, SET THE SDA.
                BTFSS   CARRY
                CALL    PULL_SDA_LOW            ;IF CARRY IS CLEAR, CLEAR THE SDA.
                CALL    I2C_CLOCK               ;LATCH THE DATA INTO THE PLL.
                DECFSZ  COUNT,F
                GOTO    TX_BIT_LOOP             ;LOOP UNTIL ALL BITS ARE SENT.
                CALL    LET_SDA_HIGH            ;LET THE SDA FLOAT HIGH.
                CALL    I2C_CLOCK               ;SEND AN I2C ACKNOWLEGE SIGNAL.
                RETURN

;-----------------------------------------------------------------------------------------

I2C_START:      BSF     SCL                     ;SEND AN I2C START SIGNAL.
                CALL    LET_SDA_HIGH 
                CALL    I2C_DELAY
                CALL    PULL_SDA_LOW
                GOTO    I2C_CLOCK2

;-----------------------------------------------------------------------------------------

I2C_STOP:       CALL    PULL_SDA_LOW            ;SEND AN I2C STOP SIGNAL.
                CALL    I2C_DELAY
                BSF     SCL
                CALL    I2C_DELAY
                CALL    LET_SDA_HIGH
                CALL    I2C_DELAY
                RETURN

;-----------------------------------------------------------------------------------------

I2C_CLOCK:      CALL    I2C_DELAY
                BSF     SCL                     ;PULSE THE SCL LINE
I2C_CLOCK2:     CALL    I2C_DELAY
                BCF     SCL                     ;         "
                RETURN

;-----------------------------------------------------------------------------------------

PULL_SDA_LOW:   BCF     SDA                     ;SDA WILL BE LOW WHEN SET AS O/P.
                BSF     RP0                     ;SELECT PAGE 1 FOR TRIS REG ACCESS.
                BCF     SDA                     ;SET SDA AS AN OUTPUT.
                BCF     RP0                     ;RETURN TO PAGE 0 FOR PORT ACCESS.
                RETURN

;-----------------------------------------------------------------------------------------

LET_SDA_HIGH:   BSF     RP0                     ;SELECT PAGE 1 FOR TRIS REG ACCESS.
                BSF     SDA                     ;SET SDA AS AN INPUT.
                BCF     RP0                     ;RETURN TO PAGE 0 FOR PORT ACCESS.
                RETURN

;-----------------------------------------------------------------------------------------

I2C_DELAY:      MOVLW   15
		MOVWF   COUNT1                  ; 15 LOOPS (3CLOCKS) = 50 uS.
I2C_LOOP:       DECFSZ  COUNT1,F
                GOTO    I2C_LOOP
                RETURN

;*****************************************************************************************
;  MISC ROUTINES.       
;                       LONG_DELAY              WAIT 100mS.
;-----------------------------------------------------------------------------------------

LONG_DELAY:     MOVLW   118
		MOVWF   COUNT1                  ; 118 LOOPS OF 845 uS = 100 mS.
                CLRF    COUNT2                  ; 256 LOOPS (3CLOCKS) x 1.1uS = 845 uS.
LD_LOOP:        DECFSZ  COUNT2,F
                GOTO    LD_LOOP
                DECFSZ  COUNT1,F
                GOTO    LD_LOOP
                RETURN

;*****************************************************************************************
;  BUTTON ROUTINES AND SET FREQUENCY
;
;*****************************************************************************************


KEY_SET:
		BSF	PLL_VALID		;SET VALID = VALID
		MOVF	PORTB,W			;READ PORTB STATUS
		ANDLW	B'00001111'		;MASK NOT VALID BIT
		MOVWF	TPORTB			;SAVE STATUS PORT
		ADDWF	PCL,F
		NOP				;0
		NOP				;1
		NOP				;2
		NOP				;3
		NOP				;4
		NOP				;5
		GOTO	NULL			;6
		GOTO	FREQ_4			;7
		NOP				;8
		NOP				;9
		GOTO	NULL			;10
		GOTO	FREQ_3			;11
		GOTO	NULL			;12
		GOTO	FREQ_2			;13 
		GOTO	FREQ_1			;14
                    
NULL:		BCF	PLL_VALID		;15  NULL        
		RETURN


FREQ_1:		MOVLW	F1224_H
		MOVWF	DIV_H
		MOVLW	F1224_L
		MOVWF	DIV_L
		GOTO	WRT_PLL
          
FREQ_2:		MOVLW	F1240_H
		MOVWF	DIV_H
		MOVLW	F1240_L
		MOVWF	DIV_L
		GOTO	WRT_PLL 

FREQ_3:		MOVLW	F1256_H
		MOVWF	DIV_H
		MOVLW	F1256_L
		MOVWF	DIV_L          
		GOTO	WRT_PLL

FREQ_4:		MOVLW	F1272_H
		MOVWF	DIV_H
		MOVLW	F1272_L
		MOVWF	DIV_L          
		GOTO	WRT_PLL

;________________________________________________________________________
  
	ORG 0x2100
	
EE_str1 DE      "Mod A/V + Pll ", 0
EE_str2 DE      " V. 1.0 ", 0
EE_str3 DE      "Quarzo 4 MHz", 0
EE_str4 DE      "  ", 0
EE_str5 DE      "C. IW2BC - 2004", 0
;-----------------------------------------------------------------------------------------

                END

